// Copyright 2021 The etcd Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package e2e

import (
	"os"
	"testing"

	"github.com/stretchr/testify/assert"
	"go.etcd.io/etcd/client/pkg/v3/testutil"
	"go.etcd.io/etcd/server/v3/verify"
)

func BeforeTest(t testing.TB) {
	SkipInShortMode(t)
	testutil.RegisterLeakDetection(t)
	os.Setenv(verify.ENV_VERIFY, verify.ENV_VERIFY_ALL_VALUE)

	path, err := os.Getwd()
	assert.NoError(t, err)
	tempDir := t.TempDir()
	assert.NoError(t, os.Chdir(tempDir))
	t.Logf("Changing working directory to: %s", tempDir)

	t.Cleanup(func() { assert.NoError(t, os.Chdir(path)) })
}
