// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package lease_test

import (
	"testing"

	"go.etcd.io/etcd/client/pkg/v3/testutil"
)

func TestMain(m *testing.M) {
	testutil.MustTestMainWithLeakDetection(m)
}
